/*

(C) 2009-2011 Mika Ilmaranta <ilmis@nullnet.fi>

License: GPLv2

*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdarg.h>
#include <syslog.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "config.h"
#include "forkexec.h"

static void makeplist(char *plist[], char *buf);
static void sigchld_hdl(int sig);

void forkexec(char *fmt, ...) {
	pid_t pid;
	va_list vl;
	char buf[BUFSIZ];
	char *argv[64];
	/* int exitval; */
	char *envp[64];
	char envbuf[BUFSIZ];
#if defined(DEBUG)
	int i;
#endif

	/* fill in buf */
	va_start(vl, fmt);
	vsprintf(buf, fmt, vl);
	va_end(vl);

	/* fill in envbuf */
	sprintf(envbuf, "LANG=%s PATH=%s TERM=%s", getenv("LANG"), getenv("PATH"), getenv("TERM"));

	pid = fork();

	if(pid == -1) {
		syslog(LOG_ERR, "forkexec: fork() failed \"%s\"", strerror(errno));
		return;
	}

	if(pid) {
		/* parent */
		if(cfg.debug >= 9) syslog(LOG_INFO, "%s: child process forked with pid: %d", __FUNCTION__, pid);
		return;
	}

	/* child */
	closelog(); /* openlog doesn't return a fd to set close on exec */

	/* make argv pointers */
	makeplist(argv, buf);

	/* make envp pointers */
	makeplist(envp, envbuf);

#if defined(DEBUG)
	for(i = 0; argv[i]; i++) {
		fprintf(stderr, "argv[%d] = \"%s\"\n", i, argv[i]);
	}
	for(i = 0; envp[i]; i++) {
		fprintf(stderr, "envp[%d] = \"%s\"\n", i, envp[i]);
	}
#endif

	execve(argv[0], argv, envp);
	syslog(LOG_ERR, "forkexec: child process failed to execute external command");
	exit(1); /* exec failed ... */
}

static void makeplist(char *plist[], char *buf)
{
	int i,ap;

	for(i = ap = 0; buf[i]; ap++) {
		plist[ap] = buf + i;
		for(; buf[i] && buf[i] != ' '; i++);
		if(buf[i]) buf[i++] = '\0';
	}
	plist[ap] = NULL;
}

/*
  Set up child signal handler to handle termination of children.
  This should be done once only for the main program.
*/
void create_sigchld_hdl(void) {
	struct sigaction act;

	memset (&act, 0, sizeof(act));
	act.sa_handler = sigchld_hdl;
	if (sigaction(SIGCHLD, &act, 0)) {
		syslog(LOG_ERR, "forkexec: failed to set up child signal handler: %s", strerror(errno));
		return;
	} else {
		if(cfg.debug >= 9) syslog(LOG_INFO, "forkexec: successfully set up child signal handler");
	}
}

/*
  SIGCHLD handler. Will be called for all children.
*/
static void sigchld_hdl(int sig)
{
	/* Wait for the dead process.
	 * We use a non-blocking call to be sure this signal handler will not
	 * block if a child was cleaned up in another part of the program. */
	int exitval;
	pid_t pid;

	if((pid = waitpid(-1, &exitval, WNOHANG)) != -1) {
		if(cfg.debug >= 9 && exitval) syslog(LOG_ERR, "%s: %s: child script with pid %d exited with non null exit value %d", __FILE__, __FUNCTION__, pid, exitval);
		else if(cfg.debug >= 9) syslog(LOG_ERR, "%s: %s: child script with pid %d exited successfully", __FILE__, __FUNCTION__, pid);
	} else {
		if(cfg.debug >= 9 && errno != ECHILD) syslog(LOG_ERR, "%s: %s: waitpid failed %s", __FILE__, __FUNCTION__, strerror(errno));
	}
}

/* EOF */
